/*
 * Copyright (c) 2003 Linuxant inc.
 *
 * NOTE: The use and distribution of this software is governed by the terms in
 * the file LICENSE, which is included in the package. You must read this and
 * agree to these terms before using or distributing this software.
 *
 */

#include "oscompat.h"
#include "osservices.h"
#include "oslinux.h"
#include "osmemory.h"
#include "testdebug.h"

#include <linux/delay.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/file.h>
#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,7) )
#include <linux/completion.h>
#endif

#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) )
static spinlock_t tqueue_lock __attribute__((unused)) = SPIN_LOCK_UNLOCKED;
#endif

static atomic_t locksHeld[NR_CPUS];

void OsObtainedLock(void)
{
    atomic_inc(&locksHeld[smp_processor_id()]);
}

void OsReleasedLock(void)
{
    atomic_dec(&locksHeld[smp_processor_id()]);
}

int OsLocksHeld(void)
{
    return atomic_read(&locksHeld[smp_processor_id()]);
}

/********************************************************************/
/* Critical Sections (Mutexes) */

typedef struct {
    spinlock_t spinlock;
    unsigned long flags;
    unsigned long nestcnt;
    int owner_cpu;
} CRIT_T;

GLOBAL HCRIT OsCriticalSectionCreate(void)
{
    CRIT_T *mutex;

    ASSERT(OsContextAllowsSleeping());
    mutex  = kmalloc(sizeof(CRIT_T), GFP_KERNEL);
    if(!mutex) {
	ASSERT(mutex);
	return NULL;
    }

    spin_lock_init(&mutex->spinlock);
    mutex->nestcnt = 0;

    return ((HCRIT)mutex);
}

GLOBAL VOID OsCriticalSectionDestroy(IN HCRIT hMutex)
{
    if ( NULL == hMutex )
    {
	ASSERT(hMutex);
	return;
    }
    kfree((CRIT_T *)hMutex);
}

GLOBAL VOID OsCriticalSectionAcquire(IN HCRIT hMutex)
{
    CRIT_T *mutex = (CRIT_T *) hMutex;
    unsigned long flags;
    int cpu;

    ASSERT(hMutex);

    if(!hMutex) /* temporary due to HCF global critical section issue */
	return;

    local_irq_save(flags);

    cpu = smp_processor_id();

    if(!spin_trylock(&mutex->spinlock)) {

	if(cpu == mutex->owner_cpu) {
	    mutex->nestcnt++;
	    local_irq_restore(flags);
	    return;
	}
	spin_lock(&mutex->spinlock);
    }

    mutex->owner_cpu = cpu;
    mutex->flags = flags;

    OsObtainedLock();

    return;
}

GLOBAL VOID OsCriticalSectionRelease(IN HCRIT hMutex)
{
    CRIT_T *mutex = (CRIT_T *) hMutex;

    ASSERT(hMutex);

    if(!hMutex) /* temporary due to HCF global critical section issue */
	return;

    if(mutex->nestcnt) {
	mutex->nestcnt--;
	return;
    }

    OsReleasedLock();

    spin_unlock_irqrestore(&mutex->spinlock, mutex->flags);
}

/********************************************************************/

typedef struct tagLS {
    spinlock_t spinlock;
    struct semaphore semaphore;
    void *owner;
    unsigned long nestcnt;
} LS, *PLS;

HLOCK OsLockCreate(void)
{
    PLS		pLock = (PLS)OsAllocate(sizeof(LS));

    if(pLock) {
	spin_lock_init(&pLock->spinlock);
	sema_init(&pLock->semaphore, 1);

	pLock->owner = NULL;
	pLock->nestcnt = 0;
    }

    //printk(KERN_DEBUG"%s: %p\n", __FUNCTION__, pLock);
    return (HLOCK)pLock;
}

void OsLockDestroy(HLOCK hLock)
{
    //printk(KERN_DEBUG"%s: %p\n", __FUNCTION__, hLock);
    if(hLock)
	OsFree((PLS)hLock);
}

void OsLockLock(HLOCK hLock)
{
    unsigned long flags;

    //printk(KERN_DEBUG"%s: down %p\n", __FUNCTION__, hLock);

    ASSERT(OsContextAllowsSleeping());
    spin_lock_irqsave(&((PLS)hLock)->spinlock, flags);
    if(down_trylock(&((PLS)hLock)->semaphore) != 0) {
	if(((PLS)hLock)->owner == current) {
	    ((PLS)hLock)->nestcnt++;
	    //printk(KERN_DEBUG"%s: ret %p (nestcnt=%lu)\n", __FUNCTION__, hLock, ((PLS)hLock)->nestcnt);
	    spin_unlock_irqrestore(&((PLS)hLock)->spinlock, flags);
	    return;
	}
	spin_unlock_irqrestore(&((PLS)hLock)->spinlock, flags);
	ASSERT(OsContextAllowsSleeping());
	down(&((PLS)hLock)->semaphore);
	spin_lock_irqsave(&((PLS)hLock)->spinlock, flags);
    }
    ((PLS)hLock)->owner = current;
    spin_unlock_irqrestore(&((PLS)hLock)->spinlock, flags);
    //printk(KERN_DEBUG"%s: ret %p\n", __FUNCTION__, hLock);
}

void OsLockUnlock(HLOCK hLock)
{
    unsigned long flags;

    //printk(KERN_DEBUG"%s: up %p\n", __FUNCTION__, hLock);
    spin_lock_irqsave(&((PLS)hLock)->spinlock, flags);
    if(((PLS)hLock)->nestcnt) {
	((PLS)hLock)->nestcnt--;
	spin_unlock_irqrestore(&((PLS)hLock)->spinlock, flags);
	return;
    }
    ((PLS)hLock)->owner = NULL;
    spin_unlock_irqrestore(&((PLS)hLock)->spinlock, flags);
    up(&((PLS)hLock)->semaphore);
}

BOOL OsLockTry(HLOCK hLock)
{
    unsigned long flags;

    spin_lock_irqsave(&((PLS)hLock)->spinlock, flags);
    if(down_trylock(&((PLS)hLock)->semaphore) != 0) {
	if(((PLS)hLock)->owner == current) {
	    ((PLS)hLock)->nestcnt++;
	    //printk(KERN_DEBUG"%s: ret %p (nestcnt=%lu) TRUE\n", __FUNCTION__, hLock, ((PLS)hLock)->nestcnt);
	    spin_unlock_irqrestore(&((PLS)hLock)->spinlock, flags);
	    return TRUE;
	}
	//printk(KERN_DEBUG"%s: ret %p FALSE\n", __FUNCTION__, hLock);
	spin_unlock_irqrestore(&((PLS)hLock)->spinlock, flags);
	return FALSE;
    }
    ((PLS)hLock)->owner = current;
    spin_unlock_irqrestore(&((PLS)hLock)->spinlock, flags);
    //printk(KERN_DEBUG"%s: ret %p TRUE\n", __FUNCTION__, hLock);
    return TRUE;
}

void OsLockTryUnlock(HLOCK hLock)
{
    //printk(KERN_DEBUG"%s: up %p\n", __FUNCTION__, hLock);
    OsLockUnlock(hLock);
}

/********************************************************************/

static spinlock_t atomic_lock __attribute__((unused)) = SPIN_LOCK_UNLOCKED;

/****************************************************************************************
  The OsAtomicCompareAndSwap function compares the value at the specified address with 
  oldVal. The value of newValue is written to the address only if oldValue and the 
  value at the address are equal. OsCompareAndSwap returns true if newValue is written 
  to the address; otherwise, it returns false. 

Params:
oldValue The value to compare at address.
newValue The value to write to address if oldValue compares true.
address The 4-byte aligned address of the data to update atomically.
Result: true if newValue was written to the address. 
 ****************************************************************************************/
BOOL OsAtomicCompareAndSwap (PVOID oldValue, PVOID newValue, PVOID* address)
{
#ifdef __HAVE_ARCH_CMPXCHG
    return (cmpxchg(address, oldValue, newValue) == oldValue);
#else
    unsigned long flags;

    spin_lock_irqsave(&atomic_lock, flags);
    if (*(PUINT32)address ==  (UINT32)oldValue)
    {
	*(PUINT32)address = (UINT32)newValue;
	spin_unlock_irqrestore(&atomic_lock, flags);
	return TRUE;
    }
    spin_unlock_irqrestore(&atomic_lock, flags);

    return FALSE;
#endif
}

INT32 OsAtomicAdd   (INT32 amount, PINT32 address)
{
    unsigned long flags;
    atomic_t *v = (atomic_t *)address;

    spin_lock_irqsave(&atomic_lock, flags);
    atomic_add(amount, v);
    amount = atomic_read(v);
    spin_unlock_irqrestore(&atomic_lock, flags);

    return amount;
}

INT32 OsAtomicIncrement (PINT32 address)
{
    unsigned long flags;
    atomic_t *v = (atomic_t *)address;
    INT32 amount;

    spin_lock_irqsave(&atomic_lock, flags);
    atomic_inc(v);
    amount = atomic_read(v);
    spin_unlock_irqrestore(&atomic_lock, flags);

    return amount;
}

INT32 OsAtomicDecrement (PINT32 address)
{
    unsigned long flags;
    atomic_t *v = (atomic_t *)address;
    INT32 amount;

    spin_lock_irqsave(&atomic_lock, flags);
    atomic_dec(v);
    amount = atomic_read(v);
    spin_unlock_irqrestore(&atomic_lock, flags);

    return amount;
}

/********************************************************************/

#define TICKS_TO_MSECS(ticks) (ticks*(1000/HZ))

static inline long MSECS_TO_TICKS(UINT32 msecs)
{
    long ticks = (msecs)/(1000/HZ);

    // round up to next tick
    if ((!ticks && msecs) || (msecs % (1000/HZ)) >= ((1000/HZ) / 2))
	ticks++;

    return ticks;
}

typedef struct TIME_OUT_INSTANCE_TYPE {
    struct tq_struct			TaskQueue;
    struct timer_list			Timer;
    UINT32				mSec;
    UINT32				bLocked;
    PFREE_FUNC				pFuncFree;
    PVOID				pRefData;
    PCBFUNC				pTimeOutCallBack;
} TIME_OUT_INSTANCE_T, *PTIME_OUT_INSTANCE_T;

// bit number for delete syncronization
#define TASK_DELETE		1
// bit number indicating that task has been queued
#define TASK_QUEUED		2

static void TimeOutHandler(unsigned long Data)
{
    PTIME_OUT_INSTANCE_T	pTimeOutInstance = (PTIME_OUT_INSTANCE_T)Data;

    if(test_bit(TASK_DELETE, (void*)&pTimeOutInstance->bLocked)) {
	return;
    }

    if(!test_and_set_bit(TASK_QUEUED, (void*)&pTimeOutInstance->bLocked)) {
	OsModuleUseCountInc();
	if (OsModemThreadSchedule(&pTimeOutInstance->TaskQueue) == 0) {
	    OsModuleUseCountDec();
	    clear_bit(TASK_QUEUED, (void*)(&pTimeOutInstance->bLocked));
	}
    }
}

static void TimerThreadFunction(void* pData)
{
    PTIME_OUT_INSTANCE_T	pTimeOutInstance = (PTIME_OUT_INSTANCE_T)pData;

    if(test_bit(TASK_DELETE, (void*)&pTimeOutInstance->bLocked)) {
	clear_bit(TASK_QUEUED, (void*)(&pTimeOutInstance->bLocked));
	OsModuleUseCountDec();
	return;
    }

    /* beware: callback functions might use FPU instructions */
    pTimeOutInstance->pTimeOutCallBack(pTimeOutInstance->pRefData);

    clear_bit(TASK_QUEUED, (void*)(&pTimeOutInstance->bLocked));

    /* callback might have set TASK_DELETE */
    if(!test_bit(TASK_DELETE, (void*)&pTimeOutInstance->bLocked))
	mod_timer(&(pTimeOutInstance -> Timer), jiffies + MSECS_TO_TICKS(pTimeOutInstance->mSec));

    OsModuleUseCountDec();
}

static DECLARE_TASK_QUEUE(tq_mdmthrd);
static DECLARE_WAIT_QUEUE_HEAD(mdmthrd_wait);
#ifdef DECLARE_COMPLETION
static DECLARE_COMPLETION(mdmthrd_exited);
#endif
static int mdmthrd_pid;

int OsModemThreadSchedule(PVOID task)
{
    int ret;

    if(mdmthrd_pid <= 0) {
	printk(KERN_DEBUG "%s: no modem thread?!\n", __FUNCTION__);
	return 0;
    }

    ret = queue_task((struct tq_struct *)task, &tq_mdmthrd);
    wake_up(&mdmthrd_wait);
    return ret;
}

#ifndef DECLARE_COMPLETION
static int mdmthrd_running;
#endif

#ifdef for_each_process
#define sigmask_lock sighand->siglock
#define recalc_sigpending_curtask(t) recalc_sigpending()
#else
#define sighand sig
#define recalc_sigpending_curtask(t) recalc_sigpending(t)
#endif

static int mdmthrd(void *startup)
{
    register struct task_struct *curtask = current;
    int i;

    lock_kernel();

    exit_mm(current);

    current->session = 1;
    current->pgrp = 1;

    snprintf(curtask->comm, sizeof(curtask->comm), "k%smodemd", CNXTTARGET);
    curtask->comm[sizeof(curtask->comm) - 1] = '\0';

    /* set highest priority */
#ifdef DEF_NICE
    curtask->nice = -20;
#else
    set_user_nice(curtask, -20);
#endif

#ifndef MAX_RT_PRIO
#define MAX_RT_PRIO 100
#endif
    curtask->policy = SCHED_FIFO;
    curtask->rt_priority = MAX_RT_PRIO-1;

    unlock_kernel();

#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) )
    flush_signals(curtask); /* must be called without spinlock */
#endif
    spin_lock_irq(&curtask->sigmask_lock);
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) )
    flush_signals(curtask); /* must be called with spinlock held */
#endif
    { // flush_signal_handlers
	struct k_sigaction *ka = &curtask->sighand->action[0];
	for (i = _NSIG ; i != 0 ; i--) {
	    ka->sa.sa_handler = SIG_DFL;
	    ka->sa.sa_flags = 0;
	    sigemptyset(&ka->sa.sa_mask);
	    ka++;
	}
    }
    sigemptyset(&curtask->blocked);
    recalc_sigpending_curtask(curtask);
    spin_unlock_irq(&curtask->sigmask_lock);

    for (i = 0; i < curtask->files->max_fds; i++ ) {
	if (curtask->files->fd[i])
	    sys_close(i);
    }

#ifdef DECLARE_COMPLETION
    complete((struct completion *)startup);
#else
    mdmthrd_running = 1;
#endif

    wait_event_interruptible(mdmthrd_wait, (run_task_queue(&tq_mdmthrd), FALSE));

#if ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) )
    flush_signals(curtask); /* must be called without spinlock */
#endif
    spin_lock_irq(&curtask->sigmask_lock);
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) )
    flush_signals(curtask); /* must be called with spinlock held */
#endif
    recalc_sigpending_curtask(curtask);
    spin_unlock_irq(&curtask->sigmask_lock);

    mdmthrd_pid = 0;
#ifdef DECLARE_COMPLETION
#if defined(complete_and_exit) || ( LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,9) )
    complete_and_exit(&mdmthrd_exited, 0);
#else
    complete(&mdmthrd_exited);
    up_and_exit(NULL, 0);
#endif
#else
    mdmthrd_running = 0;
    return 0;
#endif
}

static INT32 ntimers;

HOSTIMER OsCreatePeriodicTimer(IN UINT32 InitialTimeOut, IN PCBFUNC pTimeOutCallBack,
	IN PALLOC_FUNC pFuncAlloc, IN PFREE_FUNC pFuncFree,
	IN PVOID pRefData, OUT HTHREAD* pThreadId)
{
    PTIME_OUT_INSTANCE_T pTimeOutInstance;

    pTimeOutInstance = pFuncAlloc ? pFuncAlloc(sizeof(TIME_OUT_INSTANCE_T), pRefData) : OsAllocate(sizeof(TIME_OUT_INSTANCE_T));
    if(0 == pTimeOutInstance) {
	TRACE(T__ERROR, ("OsCreatePeriodicTimer ERROR, pTimeOutInstance is NULL???\n"));
	return 0;
    }

    INIT_TQUEUE(&pTimeOutInstance -> TaskQueue, TimerThreadFunction, pTimeOutInstance);

    pTimeOutInstance->pFuncFree = pFuncFree;
    pTimeOutInstance->pRefData = pRefData;
    pTimeOutInstance->pTimeOutCallBack = pTimeOutCallBack;
    pTimeOutInstance->bLocked = 0;
    pTimeOutInstance->mSec = InitialTimeOut;

    init_timer(&pTimeOutInstance->Timer);

    pTimeOutInstance->Timer.function = TimeOutHandler;
    pTimeOutInstance->Timer.data = (unsigned long)pTimeOutInstance;
    pTimeOutInstance->Timer.expires = jiffies;

    if(OsAtomicIncrement(&ntimers) == 1) {
#ifdef COMPLETION_INITIALIZER
	static struct completion startup = COMPLETION_INITIALIZER(startup);
#endif

	OsModuleUseCountInc();
#ifdef COMPLETION_INITIALIZER
	mdmthrd_pid = kernel_thread(mdmthrd, &startup, 0);
	wait_for_completion(&startup);
#else
	mdmthrd_running = 0;
	mdmthrd_pid = kernel_thread(mdmthrd, NULL, 0);
	{
	    int i;

	    /* wait 5 seconds maximum for modem thread to start */
	    for (i=5000; !mdmthrd_running && (i > 0); i -= 50) {
		OsSleep(50);
	    }
	    if(i <= 0) {
		printk(KERN_ERR"%s: giving up on modem thread\n", __FUNCTION__);
	    }
	}
#endif
	if(mdmthrd_pid <= 0) {
	    OsModuleUseCountDec();
	    OsAtomicDecrement(&ntimers);
	    if(pTimeOutInstance->pFuncFree)
		pTimeOutInstance->pFuncFree(pTimeOutInstance, pTimeOutInstance->pRefData);
	    else
		OsFree(pTimeOutInstance);
	    return NULL;
	}
    }

    if(InitialTimeOut) {
	add_timer(&(pTimeOutInstance->Timer));
    }

    if(pThreadId) {
	*pThreadId = (HTHREAD)mdmthrd_pid;
    }

    return (HOSTIMER)pTimeOutInstance;
}

void OsDestroyPeriodicTimer(IN HOSTIMER hTimeOut)
{
    PTIME_OUT_INSTANCE_T	pTimeOutInstance = (PTIME_OUT_INSTANCE_T)hTimeOut;
    int i;

    if(pTimeOutInstance == 0) {
	TRACE(T__ERROR, ("\nOSLIN ERROR: OsDestroyPeriodicTimer: pTimeOutInstance == NULL???\n"));
	return;
    }

    set_bit(TASK_DELETE, (void*)(&pTimeOutInstance->bLocked));

    del_timer_sync(&pTimeOutInstance->Timer);

    for(i=0; (i < 30) && test_bit(TASK_QUEUED, (void*)&pTimeOutInstance->bLocked); i++) {
	if(mdmthrd_pid <= 0) {
	    printk(KERN_WARNING "%s: modem thread died, canceling queued task(s)\n", __FUNCTION__);
	    clear_bit(TASK_QUEUED, (void*)(&pTimeOutInstance->bLocked));
	    OsModuleUseCountDec();
	    break;
	}
	schedule();
    }

    if(i == 30) {
	printk(KERN_WARNING "%s: task queue stuck?!\n", __FUNCTION__);
    }

    if(mdmthrd_pid <= 0) {
	printk(KERN_WARNING "%s: modem thread exited prematurely\n", __FUNCTION__);
    }

    if(OsAtomicDecrement(&ntimers) == 0) {
	if(mdmthrd_pid > 0) {
	    int r = kill_proc(mdmthrd_pid, SIGKILL, 1);
	    if(r) {
		printk(KERN_ERR"%s: kill_proc mdmthrd_pid=%d r=%d\n", __FUNCTION__, mdmthrd_pid, r);
	    }
#ifdef DECLARE_COMPLETION
	    wait_for_completion(&mdmthrd_exited);
#else
	    /* wait 5 seconds maximum for modem thread to terminate */
	    for (i=5000; mdmthrd_running && (i > 0); i -= 50) {
		OsSleep(50);
	    }
	    if(i <= 0) {
		printk(KERN_ERR"%s: giving up on modem thread\n", __FUNCTION__);
	    }
#endif
	    OsModuleUseCountDec();
	}
    }

    if(pTimeOutInstance->pFuncFree)
	pTimeOutInstance->pFuncFree(pTimeOutInstance, pTimeOutInstance->pRefData);
    else
	OsFree(pTimeOutInstance);
}

BOOL OsSetPeriodicTimer(IN HOSTIMER hTimeOut, IN UINT32 NewTimeOut)
{
    PTIME_OUT_INSTANCE_T	pTimeOutInstance = (PTIME_OUT_INSTANCE_T)hTimeOut;

    if(pTimeOutInstance == 0) {
	TRACE(T__ERROR, ("\nERROR!!! OsSetPeriodicTimer: pTimeOutInstance == NULL\n"));
	return FALSE;
    }

    if(NewTimeOut == pTimeOutInstance->mSec) {
	/* no change, don't touch it */
	return TRUE;
    }
    pTimeOutInstance->mSec = NewTimeOut;
    if(NewTimeOut) {
	clear_bit(TASK_DELETE, (void*)(&pTimeOutInstance->bLocked));
	mod_timer(&pTimeOutInstance -> Timer, jiffies + MSECS_TO_TICKS(pTimeOutInstance -> mSec));
	return TRUE;
    } else {
	set_bit(TASK_DELETE, (void*)(&pTimeOutInstance->bLocked));
	del_timer(&pTimeOutInstance -> Timer);
	return FALSE;
    }
}

void OsImmediateTimeOut(IN HOSTIMER hTimeOut)
{
    PTIME_OUT_INSTANCE_T    pTimeOutInstance = (PTIME_OUT_INSTANCE_T)hTimeOut;

    mod_timer(&pTimeOutInstance -> Timer, jiffies);
}

static time_t epoch = 0;

UINT32 OsGetSystemTime(void)
{
    struct timeval		timestamp;

    do_gettimeofday(&timestamp);

    return ((timestamp.tv_sec-epoch)*1000 + timestamp.tv_usec/1000);
}

void OsSleep(UINT32 ms)
{
    /* we must schedule() for sleeps >= 10ms since HCF calls us in loops
     * waiting for asynchronous events.
     */
    if(ms >= TICKS_TO_MSECS(1)) {
	UINT32 start, end;
	long timeout = MSECS_TO_TICKS(ms);
	//long ort = timeout;

	if(!OsContextAllowsSleeping()) {
	    if(ms >= 100)
	    	printk(KERN_ERR"%s(%lu): shouldn't sleep that long in this context! in_interrupt()=%u locks=%d\n", __FUNCTION__, ms, (unsigned int)in_interrupt(), OsLocksHeld());
//	    ASSERT(OsContextAllowsSleeping());
	    mdelay(ms);
	    return;
	}

	start = OsGetSystemTime();
	do {
	    set_current_state(TASK_UNINTERRUPTIBLE);
	} while ((timeout = schedule_timeout(timeout)));
	end = OsGetSystemTime() - start;
	if(end < ms) {
	    //printk(KERN_ERR"OsSleep: short timeout; wanted %ld got %ldms ort=%ld\n", ms, end, ort);
	    OsSleep(end);
	}
	//else if(end > ms) {
	//	printk(KERN_ERR"OsSleep: timeout %ldms (%ldms too much) ort=%ld\n", ms, end - ms, ort);
	//}
    } else {
	// short delay
	mdelay(ms);
    }
}

#if ! TARGET_HCF_FAMILY

typedef struct TIMER_TAG
{
    UINT32 msec;
    struct timer_list timer;
} TIMER_T, *PTIMER_T;

HANDLE OsCreateTimer(UINT32 msec, PVOID pCBFunc, PVOID pRefData)
{
    PTIMER_T pTimer;

    if ( pCBFunc == NULL )
    {
	ASSERT(pCBFunc);
	return (NULL);
    }

    ASSERT(OsContextAllowsSleeping());
    pTimer = kmalloc(sizeof(TIMER_T),GFP_KERNEL);
    if ( NULL == pTimer )
    {
	ASSERT(pTimer);
	return (NULL);
    }

    memset(pTimer,0,sizeof(TIMER_T));

    init_timer(&(pTimer->timer));
    pTimer->timer.function = pCBFunc ;
    pTimer->timer.data = (unsigned long)pRefData;
    pTimer->timer.expires = jiffies + MSECS_TO_TICKS(msec);
    pTimer->msec = msec;
    return ((HANDLE)pTimer);
}

void OsSetTimer(PVOID pTimer)
{
    struct timer_list*	pTimerH;
    PTIMER_T pTimerS = (PTIMER_T)pTimer;
    if ( pTimer == NULL )
    {
	ASSERT(pTimer);
	return;
    }
    pTimerH = &(pTimerS->timer);
    mod_timer(pTimerH, jiffies + MSECS_TO_TICKS(pTimerS->msec));
}

void OsCancelTimer(PVOID pTimer)
{
    struct timer_list*	pTimerH;
    if ( pTimer == NULL )
    {
	ASSERT(pTimer);
	return;
    }
    pTimerH = &((PTIMER_T)pTimer)->timer;

    del_timer(pTimerH);
}

void OsChangeTimerTimeOut(PVOID pTimer, UINT32 msec)
{
    PTIMER_T pTimerS = (PTIMER_T)pTimer;
    if ( pTimer == NULL )
    {
	ASSERT(pTimer);
	return;
    }
    pTimerS->msec = msec;
}

void OsDestroyTimer(PVOID pTimer)
{
    if ( pTimer == NULL )
    {
	ASSERT(pTimer);
	return;
    }

    del_timer_sync(&((PTIMER_T)pTimer)->timer);
    kfree(pTimer);
}

static DWORD dwProcessorFreq;

GLOBAL DWORD  OsGetProcessorFreq(void)
{
    return dwProcessorFreq;
}

static unsigned long OsCalcCpuRate(void)
{
    struct timeval tv;
    unsigned long flags;
    unsigned long time1, time2, cpurate;
    unsigned int target_usec;

    // first ensure that the tv_usec will not wraparound
    do_gettimeofday( &tv );

    if ( tv.tv_usec > 990000 ) {
	do
	{
	    do_gettimeofday( &tv );
	} while ( tv.tv_usec > 990000 );
    }

    local_irq_save(flags);
    // calc the cpu rate 
    do_gettimeofday( &tv );
    rdtscl(time1);
    target_usec = tv.tv_usec + 1000;
    do
    {
	do_gettimeofday( &tv );
    } while ( tv.tv_usec < target_usec );
    rdtscl(time2);

    local_irq_restore(flags);

    cpurate = ( (time2-time1+500) /1000 ) * 1000000; // to get whole number 

#ifdef DEBUG
    printk(KERN_DEBUG"%s: %lu\n", __FUNCTION__, cpurate);
#endif

    return cpurate;
}

void OsInitTime(void)
{
    struct timeval timestamp;
    do_gettimeofday(&timestamp);

    epoch = timestamp.tv_sec;

    dwProcessorFreq = OsCalcCpuRate()/1000000;
}

#endif

/********************************************************************/

HTHREAD OsGetCurrentThread(void)
{
    return (HTHREAD)((in_interrupt()) ? 0 : current->pid);
}

/********************************************************************/


void    OsErrorVPrintf       (LPCSTR szFormat, PVOID args)
{
    char printk_buf[1024];

    printk_buf[0]= '\0';
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,4,10) )
    vsprintf(printk_buf, szFormat, args);
#else
    vsnprintf(printk_buf, sizeof(printk_buf), szFormat, args);
#endif

    printk_buf[sizeof(printk_buf)-1]= '\0';

#if 0
    {
	static int lasthadnl=1;
	if(lasthadnl)
	    printk(KERN_ERR"%s", printk_buf);
	else
	    printk("%s", printk_buf);

	lasthadnl = printk_buf[strlen(printk_buf) - 1] == '\n';
    }
#else
    printk("%s", printk_buf);
#endif
}

void    OsErrorPrintf       (LPCSTR szFormat, ...)
{
    va_list args;

    va_start(args, szFormat);
    OsErrorVPrintf(szFormat, args);
    va_end(args);
}

void    OsDebugVPrintf       (LPCSTR szFormat, PVOID args)
{
    char printk_buf[1024];

    printk_buf[0]= '\0';
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,4,10) )
    vsprintf(printk_buf, szFormat, args);
#else
    vsnprintf(printk_buf, sizeof(printk_buf), szFormat, args);
#endif

    printk_buf[sizeof(printk_buf)-1]= '\0';

    printk("%s", printk_buf);
}

void    OsDebugPrintf       (LPCSTR szFormat, ...)
{
    va_list args;

    va_start(args, szFormat);
    OsDebugVPrintf(szFormat, args);
    va_end(args);
}

void    OsDebugBreakpoint   ( LPCSTR szMsg )
{
    OsDebugPrintf("%s: %s\n", __FUNCTION__, szMsg);
    KDB_ENTER();
}

#if 0
void OsDebugDumpData(PUINT8 data, UINT32 len)
{
    OsDebugPrintf("%s: data=%p len=%ld\n", __FUNCTION__, data, len);
}
#endif



typedef struct {
    wait_queue_head_t wq;
    int flag;
    spinlock_t lock;
} EVENT, *PEVENT;

HOSEVENT OsCreateEvent(void)
{
    PEVENT pEvent = (PEVENT) kmalloc(sizeof(EVENT), GFP_KERNEL);

    if(pEvent) {
	init_waitqueue_head(&pEvent->wq);
	pEvent->flag = 0;
	spin_lock_init(&pEvent->lock);
    }

    //	DPRINTF(("%s: %p", __FUNCTION__, pEvent));
    return (HOSEVENT)pEvent;
}

// Destroys event object
void OsDeleteEvent(HOSEVENT hEvent)
{
    PEVENT pEvent = (PEVENT)hEvent;

    //	DPRINTF(("%s: %p", __FUNCTION__, hEvent));
    kfree(pEvent);
}

void OsWaitForEvent(HOSEVENT hEvent)
{
#if 1 // XXX
    //	DPRINTF(("%s: %p", __FUNCTION__, hEvent));
    OsWaitForEventTime(hEvent, 10*1000);
#else
    PEVENT pEvent = (PEVENT)hEvent;
    DECLARE_WAITQUEUE(wait, current);
    unsigned long flags;

    //	DPRINTF(("%s: %p", __FUNCTION__, hEvent));

    ASSERT(OsContextAllowsSleeping());

    add_wait_queue(&pEvent->wq, &wait);
    while(1) {
	spin_lock_irqsave(&pEvent->lock, flags);
	if(pEvent->flag) {
	    pEvent->flag = 0;
	    spin_unlock_irqrestore(&pEvent->lock, flags);
	    break;
	}
	spin_unlock_irqrestore(&pEvent->lock, flags);
	rmb();
	set_current_state(TASK_UNINTERRUPTIBLE);
	schedule();
    }
    current->state = TASK_RUNNING;
    remove_wait_queue(&pEvent->wq, &wait);
#endif
}

// sets event to signalled state
void OsSetEvent(HOSEVENT hEvent)
{
    PEVENT pEvent = (PEVENT)hEvent;
    unsigned long flags;

    //	DPRINTF(("%s: %p", __FUNCTION__, hEvent));
    spin_lock_irqsave(&pEvent->lock, flags);
    pEvent->flag = 1;
    spin_unlock_irqrestore(&pEvent->lock, flags);
    wmb();
    wake_up(&pEvent->wq);
}

// waits until event is set by OsSetEvent() call or timeout (ms) has passed
WAIT_RESULT OsWaitForEventTime(HOSEVENT hEvent, UINT32 timeout)
{
    PEVENT pEvent = (PEVENT)hEvent;
    DECLARE_WAITQUEUE(wait, current);
    unsigned long flags;
    long tmticks = MSECS_TO_TICKS(timeout);
    WAIT_RESULT res = WAIT_TIMEOUT;

    //	DPRINTF(("%s: %p timeout=%ld", __FUNCTION__, hEvent, timeout));

    if(!timeout) {
	spin_lock_irqsave(&pEvent->lock, flags);
	if(pEvent->flag) {
	    pEvent->flag = 0;
	    res = WAIT_OK;
	}
	spin_unlock_irqrestore(&pEvent->lock, flags);
	return res;
    }

#ifdef DEBUG
    if(!OsContextAllowsSleeping()) {
	printk(KERN_ERR"%s(): shouldn't sleep in this context! in_interrupt()=%u locks=%d\n", __FUNCTION__, (unsigned int)in_interrupt(), OsLocksHeld());
    }
#endif
    ASSERT(OsContextAllowsSleeping());

    add_wait_queue(&pEvent->wq, &wait);
    do {
	spin_lock_irqsave(&pEvent->lock, flags);
	if(pEvent->flag) {
	    pEvent->flag = 0;
	    spin_unlock_irqrestore(&pEvent->lock, flags);
	    res = WAIT_OK;
	    break;
	}
	spin_unlock_irqrestore(&pEvent->lock, flags);
	rmb();
	set_current_state(TASK_UNINTERRUPTIBLE);
    } while ( tmticks && (tmticks = schedule_timeout(tmticks)) );

    current->state = TASK_RUNNING;
    remove_wait_queue(&pEvent->wq, &wait);

    if(res == WAIT_TIMEOUT) {
	//		DPRINTF(("%s: %p returning WAIT_TIMEOUT", __FUNCTION__, hEvent));
    }
    return res;
}

// sets event to non-signalled state
void OsClearEvent(HOSEVENT hEvent)
{
    PEVENT pEvent = (PEVENT)hEvent;
    unsigned long flags;

    //	DPRINTF(("%s: %p", __FUNCTION__, hEvent));

    spin_lock_irqsave(&pEvent->lock, flags);
    pEvent->flag = 0;
    spin_unlock_irqrestore(&pEvent->lock, flags);
    wmb();
}

#ifdef TARGET_HCF_PCI_LINUX
typedef struct tasklet_struct OSDPC, *POSDPC;

HOSDPC OsDpcCreate(void (*func)(void *), void * data)
{
    POSDPC pDpc = (POSDPC) kmalloc(sizeof(OSDPC), GFP_KERNEL);

    if(pDpc) {
	tasklet_init(pDpc, (void (*)(unsigned long))func, (unsigned long)data);
    }

    //	DPRINTF(("%s: %p", __FUNCTION__, pDpc));
    return (HOSDPC)pDpc;
}

void OsDpcDestroy(HOSDPC hDpc)
{
    POSDPC pDpc = (POSDPC)hDpc;

    tasklet_disable(pDpc);

    //	DPRINTF(("%s: %p", __FUNCTION__, hDpc));
    kfree(pDpc);
}

void OsDpcSchedule(HOSDPC hDpc)
{
    tasklet_schedule((POSDPC)hDpc);
}

void OsDpcEnable(HOSDPC hDpc)
{
    tasklet_enable((POSDPC)hDpc);
}

void OsDpcDisable(HOSDPC hDpc)
{
    tasklet_disable((POSDPC)hDpc);
}
#endif /* TARGET_HCF_PCI_LINUX */

